"""The delta module."""

from datetime import timedelta
from functools import wraps

from babel.core import default_locale

from . import parser


def _asdelta(func):
    """
    Simple decorator to convert return from timedelta.__<math>__ methods to Delta object. This is
    primarily needed because one cannot simply override the existing timedelta method due to the way
    the class is implemented. Doing so results in slotted attributes not being available to the
    subclassed methods. Therefore, we need to convert the returned timedelta to Delta instead.

    Note:
        We could reimplement all of the arithmetic magic methods but prefer not to. However, we do
        end up creating timedelta objects twice (one from the timedelta result, another when we
        create a new Delta) so there would be some performance gains from doing so though.
    """
    # NOTE: We're setting assigned because in Python 2.7, @wraps fails for certain timedelta magic
    # methods due to certain attributes missing from the timedelta class that @wraps looks for by
    # default.
    @wraps(func, assigned=("__name__", "__doc__"))
    def decorated(*args, **kwargs):
        result = func(*args, **kwargs)

        if isinstance(result, timedelta):
            return Delta.fromtimedelta(result)
        elif isinstance(result, tuple):  # pragma: no cover
            # This handles __divmod__ return.
            return tuple(
                Delta.fromtimedelta(item) if isinstance(item, timedelta) else item
                for item in result
            )
        else:  # pragma: no cover
            return result

    return decorated


def get_locale(locale=None, default="en_US_POSIX"):
    """Return default locale to use if one is not provided."""
    if not locale:
        locale = default_locale("LC_TIME")

        if not locale:
            locale = default

    return locale


class Delta(timedelta):
    """An extension of ``datetime.timedelta`` that provides additional functionality."""

    @classmethod
    def parse(cls, obj):
        """
        Return :class:`.Delta` object parsed from `obj`.

        Args:
            obj (str|number|timedelta): Object to parse into a :class:`.Delta` object.

        Returns:
            :class:`.Delta`
        """
        return cls.fromtimedelta(parser.parse_timedelta(obj))

    @classmethod
    def fromtimedelta(cls, delta):
        """
        Return :class:`.Delta` object from a native timedelta object.

        Returns:
            :class:`.Delta`
        """
        return cls(seconds=delta.total_seconds())

    def format(
        self,
        format="long",
        granularity="second",
        threshold=0.85,
        add_direction=False,
        locale=None,
    ):
        """
        Return timedelta as a formatted string.

        Args:
            format (str, optional): Can be one of "long", "short", or "narrow". Defaults to
                `"long"`.
            granularity (str, optional): The smallest unit that should be displayed. The value can
                be one of "year", "month", "week", "day", "hour", "minute" or "second". Defaults to
                `"second"`.
            threshold (float, optional): Factor that determines at which point the presentation
                switches to the next higher unit. Defaults to `0.85`.
            add_direction (bool, optional): If ``True`` the return value will include directional
                information (e.g. `'1 hour ago'`, `'in 1 hour'`). Defaults to ``False``.
            locale (str|Locale, optional): A ``Locale`` object or locale identifier. Defaults to
                system default.

        Returns:
            str
        """
        return parser.format_timedelta(
            self,
            format=format,
            granularity=granularity,
            threshold=threshold,
            add_direction=add_direction,
            locale=get_locale(locale),
        )

    def __float__(self):
        """Return class as float which returns the same as :meth:`total_seconds`."""
        return self.total_seconds()

    def __int__(self):
        """Return class as integer which returns the integer part of
        :meth:`total_seconds`."""
        return int(float(self))

    def __repr__(self):  # pragma: no cover
        """Return representation of :class:`.Delta`."""
        return f"<{self.__class__.__name__} [{self}]>"


# See _asdelta() docstring for details on why we are doing this.
Delta.__add__ = _asdelta(Delta.__add__)
Delta.__radd__ = _asdelta(Delta.__radd__)
Delta.__sub__ = _asdelta(Delta.__sub__)
Delta.__mul__ = _asdelta(Delta.__mul__)
Delta.__rmul__ = _asdelta(Delta.__rmul__)
Delta.__floordiv__ = _asdelta(Delta.__floordiv__)
Delta.__pos__ = _asdelta(Delta.__pos__)
Delta.__neg__ = _asdelta(Delta.__neg__)
Delta.__abs__ = _asdelta(Delta.__abs__)
Delta.__truediv__ = _asdelta(Delta.__truediv__)
Delta.__mod__ = _asdelta(Delta.__mod__)
Delta.__divmod__ = _asdelta(Delta.__divmod__)
# Override timedelta.min/max/resolution with equivalent Delta objects.
Delta.min = Delta(-999999999)
Delta.max = Delta(days=999999999, hours=23, minutes=59, seconds=59, microseconds=999999)
Delta.resolution = Delta(microseconds=1)


def to_seconds(*, microseconds=0, milliseconds=0, seconds=0, minutes=0, hours=0, days=0, weeks=0):
    """Return total number of seconds by combining values from each unit."""
    return Delta(
        microseconds=microseconds,
        milliseconds=milliseconds,
        seconds=seconds,
        minutes=minutes,
        hours=hours,
        days=days,
        weeks=weeks,
    ).total_seconds()



import inspect
from unittest import mock

import pytest

import zulu


parametrize = pytest.mark.parametrize


@parametrize(
    "api_func, target_func, args",
    [
        (zulu.now, "zulu.Zulu.now", ()),
        (zulu.parse, "zulu.Zulu.parse", (mock.DEFAULT,)),
        (zulu.range, "zulu.Zulu.range", (mock.DEFAULT, mock.DEFAULT, mock.DEFAULT)),
        (zulu.span_range, "zulu.Zulu.span_range", (mock.DEFAULT, mock.DEFAULT, mock.DEFAULT)),
        (zulu.parse_delta, "zulu.delta.Delta.parse", (mock.DEFAULT,)),
    ],
)
def test_api_mapping(api_func, target_func, args):
    with mock.patch(target_func) as mocked_target_func:
        result = api_func(*args)
        if inspect.isgeneratorfunction(api_func):
            tuple(result)

    assert mocked_target_func.called



from datetime import timedelta
import pickle

import pytest

from zulu import Delta, ParseError, to_seconds


parametrize = pytest.mark.parametrize


@parametrize(
    "obj,expected",
    [
        (timedelta(seconds=60), Delta(seconds=60)),
        (60, Delta(seconds=60)),
        ("32m", Delta(minutes=32)),
        ("2h32m", Delta(hours=2, minutes=32)),
        ("3d2h32m", Delta(days=3, hours=2, minutes=32)),
        ("1w3d2h32m", Delta(weeks=1, days=3, hours=2, minutes=32)),
        ("1w 3d 2h 32m", Delta(weeks=1, days=3, hours=2, minutes=32)),
        ("1 w 3 d 2 h 32 m", Delta(weeks=1, days=3, hours=2, minutes=32)),
        ("4:13", Delta(minutes=4, seconds=13)),
        ("4:13:02", Delta(hours=4, minutes=13, seconds=2)),
        ("4:13:02.266", Delta(hours=4, minutes=13, seconds=2, microseconds=266000)),
        (
            "2:04:13:02.266",
            Delta(days=2, hours=4, minutes=13, seconds=2, microseconds=266000),
        ),
        ("2 days, 4:13:02", Delta(days=2, hours=4, minutes=13, seconds=2)),
        (
            "2 days, 4:13:02.266",
            Delta(days=2, hours=4, minutes=13, seconds=2, microseconds=266000),
        ),
        ("5hr34m56s", Delta(hours=5, minutes=34, seconds=56)),
        ("5 hours, 34 minutes, 56 seconds", Delta(hours=5, minutes=34, seconds=56)),
        ("5 hrs, 34 mins, 56 secs", Delta(hours=5, minutes=34, seconds=56)),
        (
            "2 days, 5 hours, 34 minutes, 56 seconds",
            Delta(days=2, hours=5, minutes=34, seconds=56),
        ),
        ("1.2 m", Delta(seconds=72)),
        ("1.2 min", Delta(seconds=72)),
        ("1.2 mins", Delta(seconds=72)),
        ("1.2 minute", Delta(seconds=72)),
        ("1.2 minutes", Delta(seconds=72)),
        ("172 hours", Delta(days=7, seconds=14400)),
        ("172 hr", Delta(days=7, seconds=14400)),
        ("172 h", Delta(days=7, seconds=14400)),
        ("172 hrs", Delta(days=7, seconds=14400)),
        ("172 hour", Delta(days=7, seconds=14400)),
        ("1.24 days", Delta(days=1, seconds=20736)),
        ("5 d", Delta(days=5)),
        ("5 day", Delta(days=5)),
        ("5 days", Delta(days=5)),
        ("5.6 wk", Delta(days=39, seconds=17280)),
        ("5.6 week", Delta(days=39, seconds=17280)),
        ("5.6 weeks", Delta(days=39, seconds=17280)),
    ],
)
def test_delta_parse(obj, expected):
    assert Delta.parse(obj) == expected


def test_delta_as_float():
    secs = 10.1234
    delta = Delta(seconds=secs)
    assert float(delta) == secs


def test_delta_as_int():
    secs = 10.1234
    delta = Delta(seconds=secs)
    assert int(delta) == int(secs)


@parametrize("obj,exception", [({}, TypeError), ("", ParseError), ("a", ParseError)])
def test_delta_parse_invalid(obj, exception):
    with pytest.raises(exception):
        Delta.parse(obj)


@parametrize(
    "delta,opts,expected",
    [
        (Delta(seconds=5), {}, "5 seconds"),
        (Delta(seconds=5), {"format": "short"}, "5 sec"),
        (Delta(seconds=5), {"format": "narrow"}, "5s"),
        (Delta(seconds=5), {"add_direction": True}, "in 5 seconds"),
        (Delta(seconds=-5), {"add_direction": True}, "5 seconds ago"),
        (Delta(days=1, hours=10), {}, "1 day"),
        (Delta(days=1), {"granularity": "year"}, "1 year"),
        (Delta(minutes=15), {"threshold": 0.15}, "0 hours"),
        (Delta(hours=23), {"threshold": 0.85}, "1 day"),
        (Delta(hours=23), {"threshold": 1.00}, "23 hours"),
        (Delta(hours=23), {"threshold": 1.00, "locale": "de"}, "23 Stunden"),
    ],
)
def test_delta_format(delta, opts, expected):
    assert delta.format(**opts) == expected


def test_delta_format_default_locale(monkeypatch):
    for var in ("LANGUAGE", "LC_ALL", "LC_CTYPE", "LANG"):
        monkeypatch.setenv(var, "")

    assert Delta(seconds=5).format() == "5 seconds"


@parametrize(
    "opts,exception",
    [({"granularity": "invalid"}, ValueError), ({"format": "invalid"}, ValueError)],
)
def test_delta_format_exception(opts, exception):
    with pytest.raises(exception):
        Delta(1).format(**opts)


def test_delta_math_operations_return_type():
    delta = Delta(days=1, hours=1, minutes=1, seconds=1, microseconds=1)

    assert isinstance(delta + delta, Delta)
    assert isinstance(delta - delta, Delta)
    assert isinstance(delta * 1, Delta)
    assert isinstance(delta / 1, Delta)
    assert isinstance(delta // 1, Delta)
    assert isinstance(abs(delta), Delta)
    assert isinstance(+delta, Delta)
    assert isinstance(-delta, Delta)
    assert isinstance(delta % delta, Delta)
    assert isinstance(divmod(delta, delta)[1], Delta)


def test_delta_pickle():
    delta = Delta(hours=1)
    assert pickle.loads(pickle.dumps(delta)) == delta


@parametrize(
    "case,expected",
    [
        ({"microseconds": 1000}, 0.001),
        ({"milliseconds": 1000}, 1),
        ({"seconds": 5}, 5),
        ({"seconds": 65}, 65),
        ({"minutes": 10}, 600),
        ({"hours": 4}, 14400),
        ({"days": 3}, 259200),
        ({"weeks": 2}, 1209600),
        ({"microseconds": 20000, "seconds": 5, "minutes": 2}, 125.02),
        ({"milliseconds": 25300, "seconds": 5, "minutes": 2}, 150.3),
        ({"seconds": 5, "minutes": 2}, 125),
        ({"seconds": 5, "minutes": 2, "hours": 3}, 10925),
        ({"seconds": 5, "minutes": 2, "hours": 3, "days": 2}, 183725),
        ({"seconds": 5, "minutes": 2, "hours": 3, "days": 2, "weeks": 1}, 788525),
    ],
)
def test_to_seconds(case, expected):
    assert to_seconds(**case) == expected



"""Project version information."""

from pkg_resources import DistributionNotFound, get_distribution


try:
    __version__ = get_distribution(__name__.split(".")[0]).version
except DistributionNotFound:  # pragma: no cover
    # Package is not installed.
    __version__ = None



"""The zulu library."""

from .__version__ import __version__
from .api import create, now, parse, parse_delta, range, span_range
from .delta import Delta, to_seconds
from .parser import ISO8601, TIMESTAMP, ParseError
from .timer import Timer
from .zulu import Zulu



"""The timer module."""
import time


class Timer(object):
    """
    Timer that can be used to keep track of elapsed time or to check when a timeout has expired.

    Once the timer has been initialized, start the timer with :meth:`start`. Repeated calls to
    :meth:`start` will restart the timer at the current time.

    Stop the timer with :meth:`stop`. The timer can then be restarted with :meth:`start` with the
    previously elapsed time carrying over to the next timing session (similar to how a stopwatch
    works).

    Get how much time has elapsed with :meth:`elapsed`. Get whether the timer is stopped or started
    with :meth:`stopped` and :meth:`started` respectively.

    When using a timeout (i.e. countdown mode), get the remaining time with :meth:`remaining`. Call
    :meth:`done` to check whether the timeout has expired or not.

    Args:
        timeout (int|float, optional): How long, in seconds, the countdown timer should last.
    """

    __slots__ = ("timeout", "started_at", "stopped_at")

    def __init__(self, timeout=0):
        self.timeout = timeout
        self.reset()

    def __enter__(self):
        """Enter context manager and start the timer."""
        self.start()
        return self

    def __exit__(self, exc_type, exc, exc_tb):
        """Exit context manager and stop the timer."""
        self.stop()

    def reset(self):
        """Reset the timer to its initial state."""
        self.started_at = None
        self.stopped_at = None
        return self

    def start(self):
        """Start the timer."""
        if (
            self.stopped_at is not None
            and self.started_at is not None
            and self.started_at < self.stopped_at
        ):
            offset = self.stopped_at - self.started_at
        else:
            offset = 0

        self.started_at = time.time() - offset
        self.stopped_at = None
        return self

    def stop(self):
        """Stop the timer."""
        self.stopped_at = time.time()
        return self

    def started(self):
        """Return whether the timer has been started."""
        return self.started_at is not None and (
            self.stopped_at is None or self.stopped_at < self.started_at
        )

    def stopped(self):
        """Return whether the timer is stopped."""
        return self.started_at is None or (
            self.stopped_at is not None and self.stopped_at >= self.started_at
        )

    def elapsed(self):
        """Return how long the timer has been running."""
        if self.started_at is None:
            return 0
        elif self.stopped_at is not None:
            return self.stopped_at - self.started_at
        else:
            return time.time() - self.started_at

    def remaining(self):
        """Return how much time is remaining before timer runs out."""
        return self.timeout - self.elapsed()

    def done(self):
        """Return whether :attr:`timeout` has expired and the countdown is done."""
        return self.elapsed() >= self.timeout



"""The api module."""

from .delta import Delta
from .zulu import Zulu


def create(
    year=1970, month=1, day=1, hour=0, minute=0, second=0, microsecond=0, tzinfo=None, *, fold=0
):
    """
    Factory function to create a new :class:`.Zulu` datetime using the same arguments as the class.

    Args:
        year (int|dict): Date year ``1 <= year <= 9999`` or ``dict`` containing keys corresponding
            to initialization parameter names.
        month (int): Date month ``1 <= month <= 12``.
        day (int): Date day ``1 <= day <= number of days in the given month and year``
        hour (int, optional): Time hour ``0 <= hour < 24``. Defaults to ``0``.
        minute (int, optional): Time minute ``0 <= minute < 60``. Defaults to ``0``.
        second (int, optional): Time second ``0 <= second < 60``. Defaults to ``0``.
        microsecond (int, optional): Time microsecond ``0 <= microsecond < 1000000``. Defaults to
            ``0``.
        tzinfo (None|str|tzinfo, optional): Timezone information as either a ``str`` or ``tzinfo``
            subclass. If value is a ``str``, it will be converted to a ``dateutil.tz`` timezone. If
            value is ``None``, the datetime values given are assumed to in UTC. Defaults to
            ``None``.
    """
    return Zulu(
        year=year,
        month=month,
        day=day,
        hour=hour,
        minute=minute,
        second=second,
        microsecond=microsecond,
        tzinfo=tzinfo,
        fold=fold,
    )


def now():
    """Alias to :meth:`.Zulu.now`."""
    return Zulu.now()


def parse(obj, formats=None, default_tz=None):
    """
    Return :class:`.Zulu` object parsed from `obj`.

    Args:
        obj (mixed): Object to parse into a :class:`.Zulu` object.
        formats (str|list, optional): List of string formats to use when parsing. Defaults to
            ``["ISO8601", "timestamp"]``.
        default_tz (None|str|tzinfo, optional): Default timezone to use when parsed datetime
            object does not contain a timezone. Defaults to ``UTC``.

    Returns:
        :class:`.Zulu`
    """
    return Zulu.parse(obj, formats=formats, default_tz=default_tz)


def range(frame, start, end):
    """
    Yield ranges of :class:`.Zulu` instances from `start` to `end` in steps of time frame, `frame`.

    Args:
        frame (str): A time frame (e.g. year, month, day, minute, etc).
        start (datetime): The starting datetime.
        end (datetime): The ending datetime.

    Yields:
        :class:`.Zulu`: Datetime values ranging from the given start and end datetimes.
    """
    yield from Zulu.range(frame, start, end)


def span_range(frame, start, end):
    """
    Return a range of time spans from `start` to `end` in steps of time frame, `frame`.

    Args:
        frame (str): A time frame (e.g. year, month, day, minute, etc).
        start (datetime): The starting datetime.
        end (datetime): The ending datetime.

    Yields:
        tuple: 2-element tuple of Zulu time spans
    """
    yield from Zulu.span_range(frame, start, end)


def parse_delta(obj):
    """
    Return :class:`.Delta` object parsed from `obj`.

    Args:
        obj (str|number|timedelta): Object to parse into a :class:`.Delta` object.

    Returns:
        :class:`.Delta`
    """
    return Delta.parse(obj)



from contextlib import contextmanager
import time
from unittest import mock

import pytest

from zulu import Timer


parametrize = pytest.mark.parametrize


@contextmanager
def mock_time(epoch):
    with mock.patch("time.time", return_value=epoch):
        yield


def test_timer_timing():
    timer = Timer()

    assert timer.stopped()
    assert timer.elapsed() == 0

    with mock_time(1):
        timer.start()

    assert timer.started()
    assert not timer.stopped()

    with mock_time(5):
        timer.stop()

    assert not timer.started()
    assert timer.stopped()

    assert timer.elapsed() == 4


def test_timer_toggle():
    timer = Timer()

    with mock_time(5):
        timer.start()

    with mock_time(10):
        timer.stop()
        assert timer.elapsed() == 5

    with mock_time(15):
        timer.start()

    with mock_time(30):
        timer.stop()
        assert timer.elapsed() == 20


def test_timer_timeout():
    timer = Timer(timeout=10)

    with mock_time(0):
        timer.start()
        assert timer.elapsed() == 0
        assert timer.remaining() == 10
        assert not timer.done()

    with mock_time(4):
        assert timer.elapsed() == 4
        assert timer.remaining() == 6
        assert not timer.done()

    with mock_time(10):
        assert timer.elapsed() == 10
        assert timer.remaining() == 0
        assert timer.done()


def test_timer_reset():
    timer = Timer()

    timer.start()
    timer.stop()

    assert timer.started_at > 0
    assert timer.stopped_at > 0

    timer.reset()

    assert timer.started_at is None
    assert timer.stopped_at is None


def test_timer_contextmanager():
    timer = Timer()

    with timer:
        assert timer.started()

    assert timer.stopped()



"""The parser module."""

from datetime import datetime, timedelta
from itertools import groupby

from babel.dates import (
    LC_TIME,
    format_datetime as _format_datetime,
    format_timedelta as _format_timedelta,
)
from dateutil.tz import gettz, tzlocal, tzutc
import iso8601
import pytimeparse

from .helpers import NUMBER_TYPES


UTC = tzutc()
EPOCH = datetime(1970, 1, 1, tzinfo=UTC)

ISO8601 = "ISO8601"
TIMESTAMP = "timestamp"
DEFAULT_PARSE_DATETIME_FORMATS = (ISO8601, TIMESTAMP)


# Subset of Unicode date field patterns from:
# https://www.unicode.org/reports/tr35/tr35-dates.html#Date_Field_Symbol_Table
# that are supported as an alternative to Python's strptime/strftime directives. This mapping is
# used to convert the Unicode pattern (the dict keys) to the best matched strptime directive (the
# values). For values that are tuples, the first item is the directive used for string parsing while
# the second item is the directive used for string formatting. The second items are platform
# dependent and may not work on all systems.
DATE_PATTERN_TO_DIRECTIVE = {
    "yyyy": "%Y",  # Year with century
    "yy": "%y",  # Year without century
    "YYYY": "%Y",  # Year with century
    "YY": "%y",  # Year without century
    "MMMM": "%B",  # Month's full name
    "MMM": "%b",  # Month's abbreviated name
    "MM": "%m",  # Month padded
    "M": "%m",  # Month not padded
    "DDD": "%j",  # Day of the year padded
    "DD": "%j",  # Day of the year padded
    "D": "%j",  # Day of the year not padded
    "dd": "%d",  # Day of the month padded
    "d": "%d",  # Day of the month not padded
    "EEEE": "%A",  # Weekday's full name
    "EEE": "%a",  # Weekday's abbreviated name
    "EE": "%a",  # Weekday's abbreviated name
    "E": "%a",  # Weekday's abbreviated name
    "eee": "%a",  # Weekday's abbreviated name
    "ee": "%w",  # Weekday as decimal padded
    "e": "%w",  # Weekday as decimal not padded
    "HH": "%H",  # Hour-24 padded
    "H": "%H",  # Hour-24 not padded
    "hh": "%I",  # Hour-12 padded
    "h": "%I",  # Hour-12 not padded
    "mm": "%M",  # Minute padded
    "m": "%M",  # Minute not padded
    "ss": "%S",  # Second padded
    "s": "%S",  # Second not padded
    "SSSSSS": "%f",  # Microsecond padded
    "SSSSS": "%f",  # Microsecond padded
    "SSSS": "%f",  # Microsecond padded
    "SSS": "%f",  # Microsecond padded
    "SS": "%f",  # Microsecond padded
    "S": "%f",  # Microsecond not padded
    "a": "%p",  # am or pm
    "z": "%z",  # UTC offset without separator
    "Z": "%z",  # UTC offset without separator
}

TIMEDELTA_GRANULARITIES = ("second", "minute", "hour", "day", "week", "month", "year")

TIMEDELTA_FORMATS = ("long", "short", "narrow")


class ParseError(Exception):
    """Exception raised when an object cannot be parsed as a datetime."""

    pass


def parse_datetime(obj, formats=None, default_tz=None):
    """
    Attempt to parse `obj` as a ``datetime`` using  a list of `formats`. If no timezone information
    is found in `obj` and `default_tz` is set, then the naive datetime object will be shifted to the
    default timezone.

    Args:
        obj (str|datetime): Object to parse.
        formats (str|list, optional): List of string formats to use when parsing. Defaults
            to ``["ISO8601", "X"]``.
        default_tz (None|str|tzinfo, optional): Default timezone to use when parsed
            datetime object does not contain a timezone. Defaults to ``UTC``.

    Returns:
        datetime

    Raises:
        ValueError: When `default_tz` is an unrecognized timezone.
        ParseError: When `obj` can't be parsed as a datetime.
    """
    if default_tz is None:
        default_tz = UTC

    if not is_valid_timezone(default_tz):
        raise ValueError(f"Unrecognized timezone: {default_tz}")

    if is_valid_datetime(obj):
        return obj

    if formats is None:
        formats = DEFAULT_PARSE_DATETIME_FORMATS
    elif not isinstance(formats, (list, tuple)):
        formats = [formats]

    dt = _parse_datetime_formats(obj, formats)

    if dt.tzinfo is None and default_tz is not None:
        dt = dt.replace(tzinfo=get_timezone(default_tz))

    if not has_valid_timezone(dt):  # pragma: no cover
        raise ParseError("Timezone offset must be strictly between -24/+24 " "hours")

    return dt


def _parse_datetime_formats(obj, formats):
    """Parse `obj` as datetime using list of `formats`."""
    dt = None
    errors = {}

    for format in formats:
        try:
            dt = _parse_datetime_format(obj, format)
        except Exception as exc:
            errors[format] = str(exc)
            dt = None
        else:
            break

    if dt is None:
        err = ", ".join(f'"{format}" ({errors[format]})' for format in formats)
        raise ParseError(f'Value "{obj}" does not match any format in [{err}]')

    return dt


def _parse_datetime_format(obj, format):
    """Parse `obj` as datetime using `format`."""
    if format.upper() == ISO8601:
        return iso8601.parse_date(obj, default_timezone=None)
    elif format.lower() == TIMESTAMP:
        return datetime.fromtimestamp(obj, UTC)
    else:
        if "%" not in format:
            format = _date_pattern_to_directive(format)
        return datetime.strptime(obj, format)


def format_datetime(dt, format=None, tz=None, locale=LC_TIME):
    """
    Return string formatted datetime, `dt`, using format directives or pattern in `format`. If
    timezone, `tz`, is supplied, the datetime will be shifted to that timezone before being
    formatted.

    Args:
        dt (datetime): A datetime instance.
        format (str, optional): Datetime format string. Defaults to ``None`` which uses
            ISO-8601 format.
        tz (None|str|tzinfo, optional): Timezone to shift `dt` to before formatting.
        locale (str|Locale, optional): A ``Locale`` object or locale identifier.
            Defaults to system default.

    Returns:
        str
    """
    if not isinstance(dt, datetime):
        raise TypeError(
            f"zulu.parser.format()'s first argument must be a datetime, not {type(dt).__name__}"
        )  # pragma: no cover

    if format is not None and not isinstance(format, str):
        raise TypeError(
            f"zulu.parser.format()'s second argument must be a string or None,"
            f" not {type(format).__name__}"
        )  # pragma: no cover

    if not is_valid_timezone(tz):  # pragma: no cover
        raise ValueError(f"Unrecognized timezone: {tz}")

    if format is None:
        format = ISO8601

    if tz is not None:
        dt = dt.astimezone(tz)

    if format == ISO8601:
        return dt.isoformat()
    elif "%" in format:
        return dt.strftime(format)
    else:
        # Deviate from TR35 and treat "Y" as "y" when formatting. TR35 defines "Y" as 'Year in
        # "Week of Year" based calendars in which the year transition occurs on a week boundary'.
        # This has caused lots of confusion to users and is honestly not worth sticking with the
        # standard to accommodate.
        # Users should instead use %G, %V and a weekday directive (%A, %a, %w, or %u).
        format = format.replace("Y", "y")
        return _format_datetime(dt, format, locale=locale)


def _date_pattern_to_directive(format):
    """Convert date pattern format to strptime/strftime directives."""
    return "".join(
        DATE_PATTERN_TO_DIRECTIVE.get(token, token) for token in _tokenize_date_pattern(format)
    )


def _tokenize_date_pattern(format):
    """
    Return list of date pattern tokens.

    This groups tokens by repeating characters so that each set of repeating characters is a list
    item (e.g. ``'YY-MM-dd'`` becomes ``['YY', '-', 'MM', '-', 'dd']``).
    """
    return ["".join(group) for key, group in groupby(format)]


def parse_timedelta(obj):
    """
    Attempt to parse `obj` as a ``timedelta`` from a string formatted duration.

    Args:
        obj (str|number|timedelta): Object to parse.

    Returns:
        timedelta

    Raises:
        TypeError: When `obj` is not a string or timedelta.
        ParseError: When `obj` can't be parsed as a timedelta.
    """
    if isinstance(obj, timedelta):
        return obj

    is_string = isinstance(obj, str)
    is_number = isinstance(obj, NUMBER_TYPES)

    if not is_string and not is_number:
        raise TypeError(f"Expected string or number type, not {type(obj).__name__}")

    if is_string:
        seconds = pytimeparse.parse(obj)

        if seconds is None:
            raise ParseError(f'Value "{obj}" is not a recognized duration format')
    else:
        seconds = obj

    return timedelta(seconds=seconds)


def format_timedelta(
    delta,
    format="long",
    granularity="second",
    threshold=0.85,
    add_direction=False,
    locale=LC_TIME,
):
    """
    Return timedelta as a formatted string.

    Args:
        format (str, optional): Can be one of "long", "short", or "narrow". Defaults to `"long"`.
        granularity (str, optional): The smallest unit that should be displayed. The value can be
            one of "year", "month", "week", "day", "hour", "minute" or "second". Defaults to
            `"second"`.
        threshold (float, optional): Factor that determines at which point the presentation switches
            to the next higher unit. Defaults to `0.85`.
        add_direction (bool, optional): If ``True`` the return value will include directional
            information (e.g. `'1 hour ago'`, `'in 1 hour'`). Defaults to ``False``.
        locale (str|Locale, optional): A ``Locale`` object or locale identifier. Defaults to system
            default.

    Returns:
        str
    """
    if granularity not in TIMEDELTA_GRANULARITIES:
        units = ", ".join(f'"{unit}"' for unit in TIMEDELTA_GRANULARITIES)
        raise ValueError(f'Time delta granularity must be one of {units}, not "{granularity}"')

    if format not in TIMEDELTA_FORMATS:
        formats = ", ".join(f'"{format}"' for format in TIMEDELTA_FORMATS)
        raise ValueError(f'Time delta format must be one of {formats}, not "{format}"')

    return _format_timedelta(
        delta,
        granularity=granularity,
        threshold=threshold,
        add_direction=add_direction,
        format=format,
        locale=locale,
    )


def get_timezone(tz):
    """
    Coerce `tz` into a `tzinfo` compatible object.

    If ``tz == 'local'``, then the system's local timezone will be used. If `tz` is a string other
    than ``'local'``, it will be passed to ``dateutil.tz.gettz(tz)``. Otherwise, `tz` will be
    returned as-is.
    """
    if tz is None:
        tz = UTC
    elif tz == "local":
        tz = tzlocal()
    elif isinstance(tz, str):
        tz_string = tz
        tz = gettz(tz)

        if tz is None:
            raise ValueError(f"Unrecognized timezone string: {tz_string}")

    return tz


def get_timestamp(dt):
    """Return timestamp for datetime, `dt`."""
    return (dt - EPOCH).total_seconds()


def is_valid_datetime(obj):
    """
    Return whether `obj` is an instance of ``datetime`` or contains date and time attributes.

    Returns:
        bool
    """
    if isinstance(obj, datetime):
        return True
    else:
        attrs = (
            "year",
            "month",
            "day",
            "hour",
            "minute",
            "second",
            "microsecond",
            "tzinfo",
        )
        return all(hasattr(obj, attr) for attr in attrs)


def is_valid_timezone(tz):
    """
    Return whether `tz` is a valid timezone.

    Returns:
        bool
    """
    try:
        get_timezone(tz)
    except Exception:
        return False
    else:
        return True


def has_valid_timezone(dt):
    """
    Return whether `dt` has a valid timezone with a UTC offset strictly between -24/+24 hours.

    Returns:
        bool
    """
    try:
        dt.astimezone(UTC)
    except Exception:  # pragma: no cover
        return False
    else:
        return True
